# Copyright 2013 Amr Ali
# See LICENSE file for details.

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Amr Ali"
__license__ = "GPLv3+"
__email__ = "amr.ali.cc@gmail.com"

import sys
from collections import defaultdict

def public(obj):
    """A decorator to avoid retyping function/class names in __all__."""
    _all = sys.modules[obj.__module__].__dict__.setdefault('__all__', [])
    if obj.__name__ not in _all:
        _all.append(obj.__name__)
    return obj

class SingletonAbstractMeta(type):
    """A Singleton pattern abstract metaclass."""

    def __init__(cls, name, bases, attrs):
        super(SingletonAbstractMeta, cls).__init__(name, bases, attrs)
        cls._instance = defaultdict(str)

    def __call__(cls, *args, **kwargs):
        key = cls._make_key(*args, **kwargs)
        if key not in cls._instance:
            cls._instance[key] = super(SingletonAbstractMeta, cls).__call__(
                    *args, **kwargs)
        return cls._instance[key]

    @classmethod
    def _make_key(cls, *args, **kwargs):
        """A function responsible for generating a key to check against and see
        if an instance was already constructed for that key.

        This is an abstract function, it is required to override it.
        """
        from atp.excepts import ATPUnimplementedError
        raise ATPUnimplementedError("function _make_id is not implemented")

@public
class KindSingletonMeta(SingletonAbstractMeta):
    """A singleton pattern metaclass that uses the underlying class' name
    as the unique constraint.
    """

    @classmethod
    def _make_key(cls, *args, **kwargs):
        return cls.__class__

@public
class UniqueSingletonMeta(SingletonAbstractMeta):
    """A Singleton pattern metaclass that uses arguments passed to the
    underlying class constructor as the unique constraint.
    """

    @classmethod
    def _make_key(cls, *args, **kwargs):
        key = ''
        for arg in args:
            try:
                key += str(arg)
            except:
                continue
        for kwarg, value in kwargs.iteritems():
            try:
                key += str(kwarg) + str(value)
            except:
                continue
        return key




# Copyright 2013 Amr Ali
# See LICENSE file for details.

"""
An adaptive thread pool implementation that offers you a configurable **minimum**
number of threads to be always available during the application's lifetime.
This implementation adapts to the thread consumption rate by having a pool volume
monitoring thread increasing the number of workers to accommodate the rate of tasks
submitted to the pool. Workers that sleep for a configurable period of time due to
the lack of tasks will be terminated.
"""

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Amr Ali"
__license__ = "GPLv3+"
__email__ = "amr.ali.cc@gmail.com"

import logging

from .concurrency import Task, ThreadPool, GlobalThreadPool, async_call
from .version import __version__

def load_test_suite():
    from .test import suite
    return suite

logging.getLogger("atp").addHandler(logging.NullHandler())




# Copyright 2013 Amr Ali
# See LICENSE file for details.

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Amr Ali"
__license__ = "GPLv3+"
__email__ = "amr.ali.cc@gmail.com"

import sys

from atp.mixin import public

@public
def reraise(ex, message=None):
    """Reraise the exception last happened with the original traceback.

    :param ex:
        The exception to be raised instead of the original one.
    :type ex:
        A subclass of Exception.
    :param message:
        An optional message to replace the old exception's message.
    :type message:
        str
    :raises:
        ``ex``
    """
    (_, msg, tb) = sys.exc_info()
    msg = message if message else msg
    raise (ex, msg, tb)

@public
class ATPException(Exception):
    """An ATP general exception"""
    pass

@public
class ATPUnimplementedError(ATPException):
    """An unimplemented operation error"""
    pass

@public
class ATPTaskError(ATPException):
    """A task operation error"""
    pass

@public
class ATPThreadPoolError(ATPException):
    """A thread-pool operation error"""
    pass




import unittest
import test_mixin
import test_concurrency

def __make_suite(tests):
    suite = unittest.TestSuite()
    for test in tests:
        suite.addTests(unittest.TestLoader().loadTestsFromTestCase(test))
    return suite

suite = unittest.TestSuite()
suite.addTests(__make_suite(test_mixin.tests))
suite.addTests(__make_suite(test_concurrency.tests))




#!/usr/bin/env python
# Copyright 2013 Amr Ali
# See LICENSE file for details.

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Amr Ali"
__license__ = "GPLv3+"
__email__ = "amr.ali.cc@gmail.com"

# Make sure setuptools is installed
from distribute_setup import use_setuptools
use_setuptools()

import os
from setuptools import setup, find_packages
from atp import __version__

BASE_DIR = os.path.dirname(__file__)
README_PATH = os.path.join(BASE_DIR, 'README.rst')
REQS_PATH = os.path.join(BASE_DIR, 'requirements.txt')

with open(README_PATH, 'rb') as fd:
    README = fd.read()

classifiers = [
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Natural Language :: English',
    'Intended Audience :: Developers',
    'Intended Audience :: Information Technology',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    ]

setup(
        name = 'atp',
        version = __version__,
        description = 'An adaptive thread pool implementation',
        long_description = README,
        author = 'Amr Ali',
        author_email = 'amr.ali.cc@gmail.com',
        maintainer = 'Amr Ali',
        maintainer_email = 'amr.ali.cc@gmail.com',
        url = 'https://github.com/amrali/atp',
        packages = find_packages(),
        test_suite = 'atp.load_test_suite',
        license = 'GPLv3+',
        platforms = 'Posix',
        classifiers = classifiers,
    )




# Copyright 2013 Amr Ali
# See LICENSE file for details.

__author__ = "Amr Ali"
__copyright__ = "Copyright 2013 Amr Ali"
__license__ = "GPLv3+"
__email__ = "amr.ali.cc@gmail.com"

import unittest
from atp.mixin import KindSingletonMeta, UniqueSingletonMeta

class KindTmpClass(object):
    __metaclass__ = KindSingletonMeta

    def __init__(self, name):
        self.name = name

class UniqueTmpClass(object):
    __metaclass__ = UniqueSingletonMeta

    def __init__(self, name):
        self.name = name

class TestSingleton(unittest.TestCase):
    def test_kind_singleton(self):
        test1 = KindTmpClass("test1")
        test2 = KindTmpClass("test2")
        self.assertEqual(test1.name, "test1")
        self.assertEqual(test2.name, test1.name)
        self.assertEqual(test1, test2)

    def test_unique_singleton(self):
        test1 = UniqueTmpClass("test1")
        test2 = UniqueTmpClass("test2")
        self.assertEqual(test1.name, "test1")
        self.assertEqual(test2.name, "test2")
        self.assertNotEqual(test2.name, test1.name)
        self.assertNotEqual(test1, test2)

tests = [
        TestSingleton,
        ]

